/* wptWipeFile.cpp - Secure file removal
 *	Copyright (C) 2001-2005 Timo Schulz
 *	Copyright (C) 2000 Matt Gauthier
 *
 * WinPT software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 **************************************************************************
 * This code based on the sunlink.c file from the SRM project, but        *
 * it was heavily modified to work with W32 and with added GCRYPT         *
 * support for gathering random bytes.                                    *
 *                                                                        *
 * The original code was placed under the GNU Lesser Public License,      *
 * even so I decide to put this file under the GNU General Public License.*
 **************************************************************************
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>
#include <direct.h>

#include "openpgp.h"
#include "wptW32API.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptCrypto.h"

#ifdef _MSC_VER
typedef unsigned __int64 DDWORD;
#else
typedef unsigned long long DDWORD;
#endif

typedef struct {
    HANDLE fd;
    DDWORD filesize;
    DDWORD offset;
    BYTE  *buffer;
    DWORD  buffsize;
    const char *name;
    int    n_passes;
} wipe_context_s;

struct arcfour_s {
    BYTE *seed;
    int  pos;
    BYTE sbox[256];
    DWORD i;
    DWORD j;
};


void (*progress_cb) (void *, DDWORD, DDWORD);
static void *progress_cb_value = NULL;

void (*unlink_cb)(void *, const char *, int, int, int) = NULL;
static void *unlink_cb_value = NULL;

static int init_done = 0;
static struct arcfour_s rnd;


/* Add the random from buffer @buf with the length @buf_len
   into the seed @seed. The seed length (256) used to
   emulate a circular movment. */
static void
add_random (unsigned char *seed, int *seed_pos,
	    unsigned char *buf, int buf_len)
{
    int i, s_pos = *seed_pos;

    for (i=0; i < buf_len; i++)
	seed[(s_pos++ % 256)] ^= buf[i];
    *seed_pos = s_pos % 256;
}


/* Enumerate all child windows. Store their text, their thread ID, 
   their placement and their dimension. */
static BOOL CALLBACK 
child_proc (HWND h, LPARAM l)
{
    struct arcfour_s *cb =  (struct arcfour_s*)l;
    DWORD a = (DWORD)h;
    WINDOWPLACEMENT p;
    RECT r;
    char buf[256];
    int n;

    n  = GetWindowText (h, buf, 255);
    if (n > 0)
	add_random (cb->seed, &cb->pos, (BYTE*)buf, n);
    add_random (cb->seed, &cb->pos, (BYTE*)&a, 4);
    a = GetWindowThreadProcessId (h, NULL);
    add_random (cb->seed, &cb->pos, (BYTE*)&a, 4);
    GetWindowPlacement (h, &p);
    add_random (cb->seed, &cb->pos, (BYTE*)&p, sizeof (p));
    GetWindowRect (h, &r);
    add_random (cb->seed, &cb->pos, (BYTE*)&r, sizeof (r));
    return TRUE;
}


/* Initialize the seed with all kind of system variables. */
static void
init_random (unsigned char *seed)
{
    int pos=0;
    DWORD buf[16];
    int i=0;

    buf[i++] = (DWORD)GetActiveWindow ();
    buf[i++] = (DWORD)GetCapture ();
    buf[i++] = (DWORD)GetClipboardOwner ();
    buf[i++] = (DWORD)GetClipboardViewer ();
    buf[i++] = (DWORD)GetCurrentProcess ();      
    buf[i++] = (DWORD)GetCurrentProcessId ();
    buf[i++] = (DWORD)GetCurrentThread ();
    buf[i++] = (DWORD)GetDesktopWindow ();
    buf[i++] = (DWORD)GetFocus ();
    buf[i++] = (DWORD)GetMessagePos ();
    buf[i++] = (DWORD)GetOpenClipboardWindow ();
    buf[i++] = (DWORD)GetProcessHeap ();
    buf[i++] = (DWORD)GetProcessWindowStation ();
    buf[i++] = (DWORD)GetQueueStatus (QS_ALLEVENTS);
    buf[i] = (DWORD)GetTickCount ();
    add_random (seed, &pos, (BYTE*)buf, 4*i);

    {
	POINT p;
	GetCursorPos (&p);
	add_random (seed, &pos, (BYTE*)&p, sizeof (p));
	GetCaretPos (&p);
	add_random (seed, &pos, (BYTE*)&p, sizeof (p));
    }

    {
	STARTUPINFO inf;
	inf.cb = sizeof (inf);
	GetStartupInfo (&inf);
	add_random (seed, &pos, (BYTE*)&inf, sizeof (inf));
    }

    {
	MEMORYSTATUS st;
	
	st.dwLength = sizeof (st);
	GlobalMemoryStatus (&st);
	add_random (seed, &pos, (BYTE*)&st, sizeof (st));
    }

    {
	LARGE_INTEGER in;

	QueryPerformanceFrequency (&in);
	add_random (seed, &pos, (BYTE*)&in, sizeof (in));
	QueryPerformanceCounter (&in);
	add_random (seed, &pos, (BYTE*)&in, sizeof (in));
    }
    {
	rnd.seed = seed;
	rnd.pos = pos;
	EnumChildWindows (GetDesktopWindow (), child_proc, (LPARAM)&rnd);
    }

}


/* Initialize cipher with the seed as the key. */
static void
init_arcfour (struct arcfour_s *ctx, BYTE *key)
{
    BYTE t;

    ctx->i = 0;
    ctx->j = 0;
    for (ctx->i=0; ctx->i < 256; ctx->i++)
	ctx->sbox[ctx->i] = (BYTE)ctx->i;
    for (ctx->i=0; ctx->i < 256; ctx->i++) {
	ctx->j += (ctx->j+ctx->sbox[ctx->i]+key[ctx->i]);
	ctx->j &= 255;
	t = ctx->sbox[ctx->i];
	ctx->sbox[ctx->i] = ctx->sbox[ctx->j];
	ctx->sbox[ctx->j] = t;
    }
}


/* Generate a single random byte. If the cipher is not
   init, do an init first. */
static BYTE 
rnd_byte (void)
{
    struct arcfour_s *ctx = &rnd;
    BYTE t;

    if (!init_done) {
	BYTE buf[256];
	init_random (buf);
	init_arcfour (ctx, buf);
	init_done = 1;
    }

    ctx->i++; ctx->i &= 255;
    ctx->j += ctx->sbox[ctx->i]; ctx->j &= 255;
    t = ctx->sbox[ctx->i];
    ctx->sbox[ctx->i] = ctx->sbox[ctx->j];
    ctx->sbox[ctx->j] = t;
    return ctx->sbox[(ctx->sbox[ctx->i] + ctx->sbox[ctx->j]) & 255];
}


/* Generate a single alpha-num charactor. */
static char
random_char (void)
{	
    byte c = 0;

    while  (!isalnum(c))
	c = rnd_byte ();
    return c % 127; 
}


/* Use the file handle in the context to overwrite a file
   with prepared buffer contents. */
static void 
overwrite (wipe_context_s *ctx)
{
    DDWORD blocks = 0, mod = 0;
    DWORD nwritten = 0;
    LONG size_high = 0;

    blocks = ctx->filesize / ctx->buffsize;
    mod = ctx->filesize % ctx->buffsize;
    SetFilePointer (ctx->fd, 0, &size_high, FILE_BEGIN);
    while (blocks--) {
	if (!WriteFile (ctx->fd, ctx->buffer, ctx->buffsize, &nwritten, NULL))
	    break;
	ctx->offset += nwritten;
	if (unlink_cb)
	    unlink_cb (unlink_cb_value, ctx->name, 0, (unsigned)ctx->offset, 
		       (unsigned)ctx->filesize*ctx->n_passes);
    }
    if (mod) {
	WriteFile (ctx->fd, ctx->buffer, (DWORD)mod, &nwritten, NULL);
	ctx->offset += nwritten;
	if (unlink_cb)
	    unlink_cb (unlink_cb_value, ctx->name, 0, (unsigned)ctx->offset, 
		       (unsigned)ctx->filesize*ctx->n_passes);
    }
    FlushFileBuffers (ctx->fd);
    SetFilePointer (ctx->fd, 0, &size_high, FILE_BEGIN);
}


/* fill the buffer with random of the given level. */
static void
randomize_buffer (byte *buf, size_t bufsize)
{
    const int blocksize = 512;
    int blocks = bufsize / blocksize;
    int mod = bufsize % blocksize;
    int i;

    while (blocks--) {
	for (i=0; i < blocksize; i++)
	    buf[i] = rnd_byte ();
	buf += blocksize;
    }
    for (i=0; i < mod; i++)
	buf[i] = rnd_byte ();
}


/* performs a random overwrite. */
static void
overwrite_random (int npasses, wipe_context_s * ctx) 
{  	
    int i;
    
    for (i = 0; i < npasses; i++) {
	randomize_buffer (ctx->buffer, ctx->buffsize);
	overwrite (ctx);
    }
}


/* perform an overwrite with a specific byte (like 0x00). */
static void 
overwrite_byte (int byte, wipe_context_s * ctx) 
{
    memset (ctx->buffer, byte, ctx->buffsize);
    overwrite (ctx);
} /* overwrite_byte */


/* perform an overwrite with a specific byte triple (like 0x00, 0xFF, 0xAA). */
static void
overwrite_bytes (int byte1, int byte2, int byte3, wipe_context_s * ctx)
{
    DWORD i;

    memset (ctx->buffer, byte1, ctx->buffsize);
    for (i = 1; i < ctx->buffsize; i += 3) {
	ctx->buffer[i] = byte2;
	ctx->buffer[i+1] = byte3;
    }
    overwrite (ctx);
} /* overwrite_bytes */


/* For the case the file is not a regular file (this is true for
   devices or directories) this function tries to rename the file
   to random pattern and then it will be delete (without random!). */
extern "C" int 
rename_unlink (const char *path)
{
    struct stat statbuf;
    char *new_name = NULL, *p = NULL, c;  
    int i = 0, rc = 0;
    int is_dir = 0;

    if (GetFileAttributes (path) & FILE_ATTRIBUTE_DIRECTORY)
	is_dir = 1;
    
    new_name = new char[strlen (path)+15];
    if (!new_name)
	BUG (0);
    
    strcpy (new_name, path);
    p = strrchr (new_name, '\\');
    if (p != NULL) {
	p++;
	*p = '\0';
    } 
    else
	p = new_name;
    do {
	while (i < 14) {
	    c = random_char ();
	    *p = c; 
	    p++;
	    i++;
	}
	*p = '\0';
    } while (stat (new_name, &statbuf) == 0);
    
    if (rename (path, new_name) == -1) {
	rc = WPTERR_FILE_READ;
	goto leave;
    }
    if (is_dir && RemoveDirectory (new_name) == FALSE)
	rc = WPTERR_FILE_REMOVE;
    else if (!DeleteFile (new_name))
	rc = WPTERR_FILE_REMOVE;
    
leave:
    free_if_alloc (new_name);
    return rc;
}


/* return the filesize as an 64-bit integer. */
static DDWORD
GetFileSize64 (const char * path)
{
    FILE *fp = fopen (path, "r");
    if (fp) {
	struct _stati64 statbuf;
	if (_fstati64 (fileno (fp), &statbuf) == -1)
	    return (DDWORD)-1;
	fclose (fp);
	return statbuf.st_size;
    }
    return (DDWORD)-1;
}


static int
_secure_unlink (const char * path, int mode, HANDLE *r_fd)
{
    wipe_context_s ctx;
    LONG size_high = 0;
    
    if (GetFileAttributes (path) & FILE_ATTRIBUTE_DIRECTORY)
	return rename_unlink (path);
    
    memset (&ctx, 0, sizeof (ctx));
    ctx.name = path;
    ctx.buffsize = 16384;
    ctx.buffer = new byte[ctx.buffsize];
    if (!ctx.buffer)
	BUG (0);
    
    ctx.filesize = GetFileSize64 (path);
    if (!ctx.filesize) {
	free_if_alloc (ctx.buffer);
	unlink (path);
	return 0;
    }

    ctx.fd = CreateFile (path, GENERIC_WRITE, FILE_SHARE_WRITE, NULL,
			 OPEN_ALWAYS, 0, NULL);
    if (ctx.fd == INVALID_HANDLE_VALUE) {
	free_if_alloc (ctx.buffer);
	return WPTERR_FILE_OPEN;
    }
    else if (r_fd)
	*r_fd = ctx.fd;
    
    if (unlink_cb)
	unlink_cb (unlink_cb_value, ctx.name, 0, 0, 0);

    switch (mode) {
    case WIPE_MODE_FAST:
	ctx.n_passes = 1;
	overwrite_random (1, &ctx);
	break;

    case WIPE_MODE_SIMPLE:
	ctx.n_passes = 2;
	overwrite_random (2, &ctx);
	break;
        
    case WIPE_MODE_DOD:
	ctx.n_passes = 5;
	overwrite_random (1, &ctx);
	overwrite_byte ((~1) & 0xFF, &ctx);
	overwrite_random (1, &ctx);
	overwrite_byte ((~4) & 0xFF, &ctx);
	overwrite_random (1, &ctx);
	break;
        
    case WIPE_MODE_GUTMANN:
	ctx.n_passes = 39;
	overwrite_random (4, &ctx);
	overwrite_byte( 0x55, &ctx );
	overwrite_byte ( 0xAA, &ctx );
	overwrite_bytes( 0x92, 0x49, 0x24, &ctx );
	overwrite_bytes( 0x49, 0x24, 0x92, &ctx );
	overwrite_bytes( 0x24, 0x92, 0x49, &ctx );
	overwrite_byte( 0x00, &ctx );
	overwrite_byte( 0x11, &ctx );
	overwrite_byte( 0x22, &ctx );
	overwrite_byte( 0x33, &ctx );
	overwrite_byte( 0x44, &ctx );
	overwrite_byte( 0x55, &ctx );
	overwrite_byte( 0x66, &ctx );
	overwrite_byte( 0x77, &ctx );
	overwrite_byte( 0x88, &ctx );
	overwrite_byte( 0x99, &ctx );
	overwrite_byte( 0xAA, &ctx );
	overwrite_byte( 0xBB, &ctx );
	overwrite_byte( 0xCC, &ctx );
	overwrite_byte( 0xDD, &ctx );
	overwrite_byte( 0xEE, &ctx );
	overwrite_byte( 0xFF, &ctx );
	overwrite_bytes( 0x92, 0x49, 0x24, &ctx );
	overwrite_bytes( 0x49, 0x24, 0x92, &ctx );
	overwrite_bytes( 0x24, 0x92, 0x49, &ctx );
	overwrite_bytes( 0x6D, 0xB6, 0xDB, &ctx );
	overwrite_bytes( 0xB6, 0xDB, 0x6D, &ctx );
	overwrite_bytes( 0xDB, 0x6D, 0xB6, &ctx );
	overwrite_random( 4, &ctx );
	break;
    }
        
    /* Set file length to zero so allocated clusters cannot be trailed */	
    SetFilePointer (ctx.fd, 0, &size_high, FILE_BEGIN);
    SetEndOfFile (ctx.fd);
    CloseHandle (ctx.fd);
    
    memset (ctx.buffer, 0, ctx.buffsize); /* burn the last evidence */	
    free_if_alloc (ctx.buffer);

    return rename_unlink (path);
}


/* Delete a file in a secure way with the given mode @mode. */
extern "C" int 
secure_unlink (const char *path, int mode)
{
    /* If the file has one of the following attributes, the
       chance the file really gets overwritten is very low so
       we just to an unlink to spare time and entropy. */
    DWORD attr = GetFileAttributes (path);
    if ((attr & FILE_ATTRIBUTE_COMPRESSED) ||
	(attr & FILE_ATTRIBUTE_ENCRYPTED) ||
	(attr & FILE_ATTRIBUTE_SPARSE_FILE))
	return DeleteFile (path); /* XXX */
    return _secure_unlink (path, mode, NULL);
}


/* Set the callback @cb for the wipe function. The callback is call every time
   the wipe function writes data to the file. */
extern "C" void
secure_unlink_set_cb (void (*cb)(void *, const char *, int, int, int), 
		      void *cb_value)
{
    unlink_cb = cb;
    unlink_cb_value = cb_value;
}


/* wipe all free space of the given drive by creating a temp file
   which has the size of the free space. This file will be then
   overwritten with random and static pattern. */
extern "C" int
wipe_freespace (const char * drive, HANDLE *r_fd,
	        void (*cb)(void *, DDWORD, DDWORD), void * cb_value)
{
    ULARGE_INTEGER caller, total, frees;
    LONG hpart=0;
    HANDLE fd;
    int disktyp = GetDriveType (drive);
    int rc;
    char * file;

    if (disktyp != DRIVE_FIXED && disktyp != DRIVE_REMOVABLE)
	return -1;
    if (!GetDiskFreeSpaceEx (drive, &caller, &total, &frees))
	return -1;   

    /* disk is full */
    if (!frees.LowPart)
	return 0;
    file = new char[strlen (drive)+8];
    if (!file)
	BUG (0);
    sprintf (file, "%stemp_winpt.tmp", drive);
    fd = CreateFile (file,
		    GENERIC_READ|GENERIC_WRITE,
		    FILE_SHARE_READ|FILE_SHARE_WRITE,
		    NULL, CREATE_ALWAYS, 0, NULL);
    if (fd == INVALID_HANDLE_VALUE) {
	free_if_alloc (file);
	return WPTERR_FILE_OPEN;
    }
    hpart = frees.HighPart;
    SetFilePointer (fd, frees.LowPart, &hpart, FILE_BEGIN);
    SetEndOfFile (fd);
    CloseHandle (fd);

    if (cb && cb_value) {
	progress_cb = cb;
	progress_cb_value = cb_value;
    }
    rc = _secure_unlink (file, WIPE_MODE_FAST, r_fd);
    free_if_alloc (file);
    return rc;
}
